/**
 * Created by gleicher on 10/9/2015.
 * Modified by John Detter on 11/25/2016
 */

/*
 * this is the "main" file - it gets loaded last - after all the objects are loaded
 * make sure that twgl is loaded first
 * it sets up the main function to be called on window.onload
 */

var createModel = function(name, obj_content, tex_content, hardness, color)
{
    var model = new WaveObj(obj_content, name);
    model.setHardness(hardness);
    model.setGlobalColor(color, 255.0);

    if(tex_content != null)
    {
        var img = new Image();
        img.src = "data:image/png;base64," + tex_content;
        model.setTexture(img);
    }

    if(scene.addObject(model))
    {
        console.log("ERROR: FAILED TO CREATE MODEL " + name);
        return null;
    }

    grobjects.push(model);

    return model;
}

 var isValidGraphicsObject = function (object) {
    if(object.name === undefined) {
        console.log("warning: GraphicsObject missing name field");
        return false;
    }

    if(typeof object.draw !== "function" && typeof object.drawAfter !== "function") {
        console.log("warning: GraphicsObject of type " + object.name + " does not contain either a draw or drawAfter method");
        return false;
    }

    if(typeof object.center !== "function") {
        console.log("warning: GraphicsObject of type " + object.name + " does not contain a center method. ");
        return false;
    }

    if(typeof object.init !== "function") {
        console.log("warning: GraphicsObject of type " + object.name + " does not contain an init method. ");
        return false;
    }

    return true;
 }
window.onload = function() {
    "use strict";

    // set up the canvas and context
    var canvas = document.createElement("canvas");
    canvas.setAttribute("width", 700);
    canvas.setAttribute("height",700);
    document.body.appendChild(canvas);

    // make a place to put the drawing controls - a div
    var controls = document.createElement("DIV");
    controls.id = "controls";
    document.body.appendChild(controls);

    // a switch between camera modes
    uiMode = document.createElement("select");
    uiMode.innerHTML += "<option>ArcBall</option>";
    uiMode.innerHTML += "<option>Drive</option>";
    uiMode.innerHTML += "<option>Fly</option>";
    uiMode.innerHTML += "</select>";
    controls.appendChild(uiMode);

    var resetButton = document.createElement("button");
    resetButton.innerHTML = "Reset View";
    resetButton.onclick = function() {
        // note - this knows about arcball (defined later) since arcball is lifted
        arcball.reset();

        drivePos = [0,.2,5];
        driveTheta = 0;
        driveXTheta = 0;

    }
    controls.appendChild(resetButton);

    // make some checkboxes - using my cheesy panels code
    checkboxes = makeCheckBoxes([ ["Run",1], ["Examine",0] ]); //


    // make some sliders - using my cheesy panels code
    sliders = makeSliders([["TimeOfDay",0,24,12]]);

    // this could be gl = canvas.getContext("webgl");
    // but twgl is more robust
    // gl = twgl.getWebGLContext(canvas);
    gl = canvas.getContext("webgl", {antialias:true});

    // make a fake drawing state for the object initialization
    var drawingState = {
        gl : gl,
        proj : twgl.m4.identity(),
        view : twgl.m4.identity(),
        camera : twgl.m4.identity(),
        sunDirection : [0,1,0]
    }
    

    // information for the cameras
    lookAt = [0,0,0];
    lookFrom = [0,5,-10];
    fov = 1.1;

    arcball = new ArcBall(canvas);

    // for timing
    realtime = 0;
    lastTime = Date.now();

    // parameters for driving
    drivePos = [0,.2,5];
    driveTheta = 0;
    driveXTheta = 0;

    // cheesy keyboard handling
    keysdown = {};

    document.body.onkeydown = function(e) {
        var event = window.event ? window.event : e;
        keysdown[event.keyCode] = true;
        e.stopPropagation();
    };
    document.body.onkeyup = function(e) {
        var event = window.event ? window.event : e;
        delete keysdown[event.keyCode];
        e.stopPropagation();
    };

    var error = false;
    console.log("Setting up scene...");
    scene = new Scene(town_vshader_content, town_fshader_content);

    /* Make sure the shaders are compiled and linked and are valid */
    if(!error && (error |= scene.compileShaders()))
        console.log("Couldn't compile shaders.");
    if(!error && (error |= scene.linkShaders()))
        console.log("Couldn't link shaders.");
    console.log("Shaders compiled and linked successfully.");
    
    /* Set the default scene properties */
    if(!error)
    {
        scene.enableDiffuse();
        // scene.disableDiffuse();
        scene.enableSpecular();
        scene.enableAmbient();
        scene.disableLightAverage();
        // scene.enableLightAverage();
        scene.denyAlpha();
        scene.enableNormals();
    }

    /* Set uniform names for other properties */
    if(!error)
        console.log("Setting other uniform properties...");
    if(!error && (error |= scene.initGLLighting(
                    "lightingProperties", "lightPositions", "lightColors",
                    "lightIntensities", "lightFalloffs", "lightTypes",
                    "cameraPosition", "lightTx")))
        console.log("\tCouldn't initialize uniforms!! FAILURE");

    if(!error && scene.setViewUniforms("view", "proj", "cameraPosition"))
    {
        console.log("\tFAILURE: Couldn't set view, proj and camera position uniforms!");
        error = true;
    }

    if(!error && scene.setTextureUniforms("useTexture", "textureSample", "hardness"))
    {
        console.log("\tFAILURE: Couldn't set texture uniforms!");
        error = true;
    }

    if(!error && !scene.verifyGLUniforms()) 
        console.log("\tUniforms seem to be okay.");
    else {
        error = true;
        console.log("\tFAILURE: Please check uniform setup. Something isn't right.");
    }

    /* Set the attribute names */
    scene.setAttributeNames("vertices", "normals", "diffuse", "texCoord", "model");

    /* Add lights */
    if(!error) console.log("Adding lights...");
    if(!error && scene.addLight(
                [0.0, 0.0, 0.0],  // position
                [1.0, 1.0, 1.0, 1.0], // color
                .3, 200.0, "ambient"))
    {
        error = true;
        console.log("\tFAILURE");
    }
    if(!error && scene.addLight(
                //[0.0, 6.0, 0.0],  // position
                [0.0, 8.0, 0.0],  // position
                [1.0, 1.0, 1.0, 1.0], // color
                .5, // intensity
                5, 
                "point"))
    {
        error = true;
        console.log("\tFAILURE");
    }
    scene.setLightAnimationCallback(1, sun_light_animation);
    if(!error)
        console.log("\tAll lights added successfully.");

    /* Load and create all models here */
    // var easy_model = createModel("easy", easy_content, easy_image, 1.0, "green");
    // var hard_model = createModel("hard", hard_content, hard_image, 1.0, "green");
    
    var plants_model = createModel("plants", plants_content, plants_image, 1.0, "green");
    plants_model.setAnimationCallback(plants_animation);
    plants_model.tomatos = createModel("tomatos", tomato_content, tomato_image, 1.0, "red");
    plants_model.tomatos.hide();
    var bricks_model = createModel("bricks", bricks_content, bricks_image, 1.0, "green");
    var doors_model = createModel("doors", doors_content, doors_image, 1.0, "green");
    var fences_model = createModel("fences", fences_content, fences_image, 1.0, "green");
    var fields_model = createModel("fields", fields_content, fields_image, 1.0, "green");
    var floors_model = createModel("floors", floors_content, floors_image, 1.0, "green");
    var ground_model = createModel("ground", ground_content, ground_image, 1.0, "green");
    var roads_model = createModel("roads", roads_content, roads_image, 1.0, "green");
    var roofs_model = createModel("roofs", roofs_content, roofs_image, 1.0, "green");
    var windmillbase_model = createModel("windmillbase", windmillbase_content, windmillbase_image, 1.0, "green");

    var windmillblades1_model = createModel("windmillblades1", windmillblades_content, windmillblades_image, 1.0, "green");
    windmillblades1_model.rotate([0.0, 215.301, 0.0]);
    windmillblades1_model.setTranslation([0.68973, 0.51205, -0.21699]);
    windmillblades1_model.setAnimationCallback(windmill1_animation);

    var windmillblades2_model = createModel("windmillblades2", windmillblades_content, windmillblades_image, 1.0, "green");
    windmillblades2_model.rotate([0.0, 35.301, 0.0]);
    windmillblades2_model.setTranslation([-0.20472, 0.51205, 1.41534]);
    windmillblades2_model.setAnimationCallback(windmill2_animation);

    // var windmillblades2_model = createModel("windmillblades", windmillblades_content, windmillblades_image, 1.0, "green");
    var windmillmain_model = createModel("windmillmain", windmillmain_content, windmillmain_image, 1.0, "green");
    var windows_model = createModel("windows", windows_content, windows_image, 1.0, "green");

    console.log("Primitives: " + scene.getPrimitiveCount());

    // a selector for which object should be examined
    toExamine = document.createElement("select");
    grobjects.forEach(function(obj) {
           toExamine.innerHTML +=  "<option>" + obj.name + "</option>";
        });
    controls.appendChild(toExamine);

    draw();
};

// the actual draw function - which is the main "loop"
function draw() {
    // advance the clock appropriately (unless its stopped)
    var curTime = Date.now();
    var timechg = 0;
    if (checkboxes.Run.checked) {
        realtime += (curTime - lastTime);
        timechg = curTime - lastTime;
        animation_time += timechg;
    }
    lastTime = curTime;


    // figure out the transforms
    var projM = twgl.m4.perspective(fov, 1, 0.1, 100);
    var cameraM = twgl.m4.lookAt(lookFrom,lookAt,[0,1,0]);
    var viewM = twgl.m4.inverse(cameraM);

    // implement the camera UI
    if (uiMode.value == "ArcBall") {
        viewM = arcball.getMatrix();
        twgl.m4.setTranslation(viewM, [0, 0, -10], viewM);
    } else if (uiMode.value == "Drive") {
        if (keysdown[65]) { driveTheta += .02; }
        if (keysdown[68]) { driveTheta -= .02; }
        if (keysdown[87]) {
            var dz = Math.cos(driveTheta);
            var dx = Math.sin(driveTheta);
            drivePos[0] -= .05*dx;
            drivePos[2] -= .05*dz;
        }
        if (keysdown[83]) {
            var dz = Math.cos(driveTheta);
            var dx = Math.sin(driveTheta);
            drivePos[0] += .05*dx;
            drivePos[2] += .05*dz;
        }

        cameraM = twgl.m4.rotationY(driveTheta);
        twgl.m4.setTranslation(cameraM, drivePos, cameraM);
        viewM = twgl.m4.inverse(cameraM);
    }else if (uiMode.value == "Fly") {

        if (keysdown[65] || keysdown[37]) { 
            driveTheta += .02; 
        }else if (keysdown[68] || keysdown[39]) { 
            driveTheta -= .02; 
        }

        if (keysdown[38]) { driveXTheta += .02; }
        if (keysdown[40]) { driveXTheta -= .02; }

        var dz = Math.cos(driveTheta);
        var dx = Math.sin(driveTheta);
        var dy = Math.sin(driveXTheta);

        if (keysdown[87]) {
            drivePos[0] -= .25*dx;
            drivePos[2] -= .25*dz;
            drivePos[1] += .25 * dy;
        }

        if (keysdown[83]) {
            drivePos[0] += .25*dx;
            drivePos[2] += .25*dz;
            drivePos[1] -= .25 * dy;
        }

        cameraM = twgl.m4.rotationX(driveXTheta);
        twgl.m4.multiply(cameraM, twgl.m4.rotationY(driveTheta), cameraM);
        twgl.m4.setTranslation(cameraM, drivePos, cameraM);
        viewM = twgl.m4.inverse(cameraM);
    }

    // get lighting information
    var tod = Number(sliders.TimeOfDay.value);
    var sunAngle = Math.PI * (tod-6)/12;
    var sunDirection = [Math.cos(sunAngle),Math.sin(sunAngle),0];

    // first, let's clear the screen
    var color_diff = Math.abs((tod - 12.0) / 48.0);
    gl.clearColor(0.3 + color_diff, 0.3, .8 - color_diff, 1.0);
    gl.enable(gl.DEPTH_TEST);
    gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

    // make a real drawing state for drawing
    var drawingState = {
        gl : gl,
        proj : projM,   // twgl.m4.identity(),
        view : viewM,   // twgl.m4.identity(),
        camera : cameraM,
        timeOfDay : tod,
        sunDirection : sunDirection,
        realtime : realtime
    }

    /* Inform our scene about changes to the environment */
    if(scene.setView(viewM, projM, lookFrom))
        console.log("FAILURE: Failed to set scene view!!");

    // initialize all of the objects that haven't yet been initialized (that way objects can be added at any point)
    grobjects.forEach(function(obj) { 
        if(!obj.__initialized) {
            if(isValidGraphicsObject(obj)){
                obj.init(drawingState);
                obj.__initialized = true;
            }
        }
    });

    // now draw all of the objects - unless we're in examine mode
    if (checkboxes.Examine.checked) {
        // get the examined object - too bad this is an array not an object
        var examined = undefined;
        grobjects.forEach(function(obj) { if (obj.name == toExamine.value) {examined=obj;}});
        var ctr = examined.center(drawingState);
        var shift = twgl.m4.translation([-ctr[0],-ctr[1],-ctr[2]]);
        twgl.m4.multiply(shift,drawingState.view,drawingState.view);
        
        scene.draw(true);
        examined.animate([timechg, 0]);
        examined.draw();
    } else {
        scene.animate([timechg, tod]);
        scene.draw(false);
    }

    window.requestAnimationFrame(draw);
};
