var town_fshader_content = `
precision highp float;

/* Transforms that need to be applied */
uniform mat4 view;
uniform mat4 proj;

/* Lighting meta information (specular) */
uniform vec3 cameraPosition;
uniform float lightingProperties[8];

/* Specific light information */
uniform vec3 lightPositions[32]; /* Positions of the lights */
uniform vec4 lightColors[32]; /* Colors of each of the lights */
uniform float lightIntensities[32]; /* Light intensities */
uniform float lightFalloffs[32]; /* Light intensities */
uniform float lightTypes[32]; /* Type of light (disabled, ambient, point) */
uniform mat4 lightTx[32]; /* Transform for the lights */

/* Texture information */
uniform float useTexture;
uniform sampler2D textureSample;

/* Material information */
uniform float hardness; /* Hardness of the object (specular) */

/* Stuff that must be passed from vertex shader */
varying vec3 fPosition;
varying vec3 fNormal;
varying vec2 fTexCoord;
varying vec4 fDiffuse;

void main()
{
    /* DEBUG */
    // vec4 c = vec4(fCameraPosition, 1.0);
    // c /= 20.0;
    // c.x = min(max(0.0, c.x), 1.0);
    // c.y = min(max(0.0, c.y), 1.0);
    // c.z = min(max(0.0, c.z), 1.0);
    // c.a = 1.0;
    // gl_FragColor = c;
    // return;

    // gl_FragColor = texture2D(textureSample, fTexCoord).rgba;
    // return;

    // gl_FragColor = c;
    // return;
    // if(lightingProperties[5] == 0.0)
    // gl_FragColor = vec4(0.0, 1.0, 0.0, 1.0);
    // else gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);
    // return;

    // gl_FragColor = vec4(abs(fNormal), 1.0);
    // return;

    // gl_FragColor = vec4(fPosition, 1.0);
    // return;

    // vec3 lightP = (lightTx[1] * vec4(lightPositions[1], 1.0)).xyz;
    // gl_FragColor = vec4(lightP, 1.0);
    // return;

    /* Some uniform lighting properties */
    float useDiffuse = lightingProperties[0];
    float useSpecular = lightingProperties[1];
    float useAmbient = lightingProperties[2];
    float useNormals = lightingProperties[3];
    float alphaAllowed = lightingProperties[4];
    float useAverage = lightingProperties[5];

    /* The running total for light results */
    vec4 diffuseResult = vec4(0.0, 0.0, 0.0, 1.0);
    vec4 specularResult = vec4(0.0, 0.0, 0.0, 1.0);
    vec4 ambientResult = vec4(0.0, 0.0, 0.0, 1.0);

    /* Calculate the camera position in the world */
    mat3 rotMat;
    rotMat[0] = view[0].xyz;
    rotMat[1] = view[1].xyz;
    rotMat[2] = view[2].xyz;

    vec3 transMat = view[3].xyz;

    /* Calculate the camera position */
    vec3 fCameraPosition = cameraPosition;
    fCameraPosition = -transMat * rotMat;

    /* Normalized vector pointing from the vertex position to the camera */
    vec3 toCamera = normalize(fCameraPosition - fPosition);

    /* Calculate the diffuse color to use for light calculations */
    vec4 diffuseColor = vec4(0.0, 0.0, 0.0, 0.0);
    if(useTexture > 0.0)
    {
        diffuseColor = texture2D(textureSample, fTexCoord).rgba;
    } else diffuseColor = fDiffuse;

    /* How many lights did we use for calculations */
    int validLights = 0;

    /* Do calculations on a per light basis */
    for(int x = 0;x < 32;x++)
    {
        /* Calculate the position of this light */
        vec3 lightPosition = (lightTx[x] * vec4(lightPositions[x], 1.0)).xyz;
        /* The normalized vector pointing from the vector to the light */
        vec3 toLight = normalize(lightPosition - fPosition);
        toLight.y = -toLight.y;

        /* Special lighting vector calculations */
        vec3 V = toCamera;
        vec3 I = toLight;
        vec3 N = normalize(fNormal);
        vec3 R = I - (2.0 * N * dot(N, I));
        float P = hardness;
        P = 1.0;
        vec4 M = lightColors[x];
        toLight.y = -toLight.y;

        if(lightTypes[x] > 0.0)
        {
            validLights++;

            if(lightTypes[x] == 1.0)
            {
                /* Ambient Light */
                ambientResult += diffuseColor * lightIntensities[x];
            } else if(lightTypes[x] == 2.0 && useNormals > 0.0)
            {   
                /* Calculate the light falloff value */
                float d = distance(lightPosition, fPosition);
                if(d < 0.1)
                    d = 0.1;
                float fallI = 0.0;

                /* Calculate light falloff value */
                fallI = ((lightFalloffs[x] * 2.0) / d);
                if(fallI <= 0.0)
                    fallI = 0.0;
                else if(fallI > 1.0)
                    fallI = 1.0;

                /* Point light */
                diffuseResult += 
                    diffuseColor
                    * fallI
                    * lightIntensities[x]
                    * max(0.0, dot(toLight, fNormal));
                ;

                float diffuseMax = 1.0;
                diffuseResult.r = max(0.0, min(diffuseResult.r, diffuseMax));
                diffuseResult.g = max(0.0, min(diffuseResult.g, diffuseMax));
                diffuseResult.b = max(0.0, min(diffuseResult.b, diffuseMax));
                diffuseResult.a = 1.0;

                float specularIntensity = 0.4;
                specularResult += 
                    specularIntensity * 
                    fallI * 
                    M * (pow(dot(R, V), P));

                float specMax = 1.0;
                specularResult.r = max(0.0, min(specularResult.r, specMax));
                specularResult.g = max(0.0, min(specularResult.g, specMax));
                specularResult.b = max(0.0, min(specularResult.b, specMax));
                specularResult.a = 1.0;
            }
        } else break;
    }

    /* Use only the values that are enabled */
    vec4 composite = vec4(0.0, 0.0, 0.0, 0.0);
    if(useAmbient > 0.0)
        composite += ambientResult;
    if(useDiffuse > 0.0)
        composite += diffuseResult;
    if(useSpecular > 0.0)
        composite += specularResult;

    /* Take the average of all lighting properties */
    // if(validLights > 0 && useAverage > 0.0)
    // composite /= float(validLights);

    composite.a = 1.0; /// hack?
    gl_FragColor = composite;
}
`;
