var town_vshader_content = `
precision highp float;

/* Transforms that need to be applied */
uniform mat4 view;
uniform mat4 proj;
uniform mat4 model;

/* Lighting meta information (specular) */
uniform vec3 cameraPosition;
uniform float lightingProperties[8];

/* Specific light information */
uniform vec3 lightPositions[32]; /* Positions of the lights */
uniform vec4 lightColors[32]; /* Colors of each of the lights */
uniform float lightIntensities[32]; /* Light intensities */
uniform float lightFalloffs[32]; /* Light intensities */
uniform float lightTypes[32]; /* Type of light (disabled, ambient, point) */
uniform mat4 lightTx[32]; /* Transform for the light position */

uniform float hardness; /* Hardness of the object (specular) */

/* Texture information */
uniform float useTexture;
uniform sampler2D textureSample;

/* Vertex attributes */
attribute vec3 vertices; /* Only actual reqired attribute */
attribute vec3 normals; /* Required for lighting */
attribute vec4 diffuse; /* Material diffuse for the object */
attribute vec2 texCoord; /* Texture coordinates (optinal) */

/* Stuff that must be passed to frag shader */
varying vec3 fPosition;
varying vec3 fNormal;
varying vec2 fTexCoord;
varying vec4 fDiffuse;

void main()
{
    /* Calculate the position of this vertex */
    vec4 pos = proj * view * model * vec4(vertices, 1.0);
    vec4 worldPos = model * vec4(vertices, 1.0);

    /* Make sure the normal is unit length */
    // fNormal = normalize(proj * view * model * vec4(normals, 0.0)).xyz;
    fNormal = normalize(model * vec4(normals, 0.0)).xyz;

    /* Set other varying properties */
    fDiffuse = diffuse;
    fPosition = worldPos.xyz;
    fTexCoord = texCoord;

    /* Set GL Position */
    gl_Position = pos;
}
`;
